# from neo4j import GraphDatabase

# # URI examples: "neo4j://localhost", "neo4j+s://xxx.databases.neo4j.io"

# import pandas as pd
# from neo4j import GraphDatabase

# class Neo4jConnection:
#     def __init__(self, uri, user, password):
#         self.__uri = uri
#         self.__user = user
#         self.__password = password
#         self.__driver = None
#         try:
#             self.__driver = GraphDatabase.driver(self.__uri, auth=(user, password))
#         except Exception as e:
#             print("Failed to create the driver:", e)
        
#     def close(self):
#         if self.__driver is not None:
#             self.__driver.close()
    
#     def execute_query(self, query, parameters=None):
#         if self.__driver is None:
#             print("Driver not initialized!")
#             return None
#         session = None
#         response = None
#         try:
#             session = self.__driver.session()
#             response = session.write_transaction(self._execute_tx, query, parameters)
#         except Exception as e:
#             print("Query failed:", e)
#         finally:
#             if session is not None:
#                 session.close()
#         return response
    
#     @staticmethod
#     def _execute_tx(tx, query, parameters):
#         result = tx.run(query, parameters)
#         return result.single()

# # Configuration for Neo4j
# neo4j_uri = "neo4j+s://6b6151d7.databases.neo4j.io"  # Change this to your Neo4j instance
# neo4j_user = "neo4j"  # Change to your Neo4j username
# neo4j_password = "m0PKWfVRYrhDUSQsTCqOGBYoGQLmN4d4gkTiOV0r8AE"  # Change to your Neo4j password

# # Initialize Neo4j connection
# neo4j_conn = Neo4jConnection(neo4j_uri, neo4j_user, neo4j_password)

# # Load CSV file
# csv_file_path = '/media/ansh/New_Volume_1/reservoir-gpt/eagle-ford.csv'  # Update this to your CSV file path
# df = pd.read_csv(csv_file_path)
# counter = 1
# # Iterate over the DataFrame and insert data into Neo4j
# for index, row in df.iterrows():
#     print("Counter", counter)
#     counter +=1
#     query = f"""
#     MERGE (n1:`{row['subject_type']}` {{name: $entity1}})
#     MERGE (n2:`{row['object_type']}` {{name: $entity2}})
#     MERGE (n1)-[:`{row['predicate']}` {{
#         sentence: $sentence, 
#         document_id: $document_id, 
#         predicate_type: $predicate_type
#     }}]->(n2)
#     """
#     parameters = {
#         'entity1': row['subject'],
#         'entity2': row['object'],
#         'sentence': row['sentence'],
#         'document_id': row['document_id'],
#         'predicate_type': row['predicate_type']
#     }
#     neo4j_conn.execute_query(query, parameters)

# # Close the Neo4j connection
# neo4j_conn.close()



# # -----------------------Upload-------------------------------------

# import boto3

# # Your AWS credentials and the S3 bucket details
# aws_access_key_id = 'AKIA5ZFZH6CA6LDWIPV5'
# aws_secret_access_key = 'wdlGk5xuwEukpN6tigXV0S+CMJKdyQse2BgYjw9o'
# bucket_name = 'pstreamsbucket1'
# file_path = 'english.pdf'  # The local path to the file
# file_key = 'english.pdf'  # The key under which to store the file in S3

# # Create an S3 client
# s3 = boto3.client(
#     's3',
#     aws_access_key_id=aws_access_key_id,
#     aws_secret_access_key=aws_secret_access_key,
#     # If you're using a specific region, uncomment and set the region_name parameter
#     # region_name='your-bucket-region'
# )

# # Upload the file
# with open(file_path, 'rb') as file:
#     s3.upload_fileobj(
#         Fileobj=file,
#         Bucket=bucket_name,
#         Key=file_key,
#         ExtraArgs={'ContentType': 'application/pdf'}
#     )

# print(f"File {file_path} uploaded to {bucket_name}/{file_key}")


# # ----------------------------------Delete--------------------------------------
# import boto3

# # Your AWS access key ID and secret access key
# aws_access_key_id = 'AKIA5ZFZH6CA6LDWIPV5'
# aws_secret_access_key = 'wdlGk5xuwEukpN6tigXV0S+CMJKdyQse2BgYjw9o'

# # The name of the bucket and the key (file name) you want to delete
# bucket_name = 'pstreamsbucket1'
# file_key = 'sample123.pdf'

# # Create an S3 client
# s3 = boto3.client(
#     's3',
#     aws_access_key_id=aws_access_key_id,
#     aws_secret_access_key=aws_secret_access_key,
#     # If you're using a specific region, uncomment and set the region_name parameter
#     # region_name='your-bucket-region'
# )

# # Function to delete a file from the specified S3 bucket
# def delete_file(bucket, key):
#     try:
#         response = s3.delete_object(Bucket=bucket, Key=key)
#         print(f"File {key} deleted from bucket {bucket}.")
#     except Exception as e:
#         print(e)

# delete_file(bucket_name, file_key)


# ------------------------------------List --------------------------------------
# import boto3

# # Your AWS access key ID and secret access key
# aws_access_key_id = 'AKIA5ZFZH6CA6LDWIPV5'
# aws_secret_access_key = 'wdlGk5xuwEukpN6tigXV0S+CMJKdyQse2BgYjw9o'

# # The name of the bucket you want to list files from
# bucket_name = 'pstreamsbucket1'

# # Create an S3 client
# s3 = boto3.client(
#     's3',
#     aws_access_key_id=aws_access_key_id,
#     aws_secret_access_key=aws_secret_access_key,
#     # If you're using a specific region, uncomment and set the region_name parameter
#     # region_name='your-bucket-region'
# )

# # List files in the specified S3 bucket
# def list_files(bucket):
#     try:
#         response = s3.list_objects_v2(Bucket=bucket)
#         if 'Contents' in response:
#             for file in response['Contents']:
#                 print(file['Key'])
#         else:
#             print("No files found.")
#     except Exception as e:
#         print(e)

# list_files(bucket_name)



import re

from querent.processors.async_processor import AsyncProcessor
from querent.logging.logger import setup_logger


class TextCleanupProcessor(AsyncProcessor):
    def __init__(self):
        self.logger = setup_logger(__name__, "TextCleanupProcessor")

    async def process_text(self, data: str) -> str:
        data = data.replace("\"", "").replace('\“', '').replace('\”', '')
        data = data.replace("\\n", " ").replace("\\t", " ")
        data = re.sub(r'\\x[0-9a-fA-F]{2}', '', data)
        data = data.replace("\n", " ").replace("\r", " ").replace("\t", " ")

        return data


import json
from unidecode import unidecode
from transformers import AutoTokenizer
from querent.kg.ner_helperfunctions.fixed_predicate import FixedPredicateExtractor
from querent.common.types.ingested_images import IngestedImages
from querent.config.core.opensource_llm_config import Opensource_LLM_Config
from querent.core.transformers.relationship_extraction_llm import RelationExtractor
from querent.kg.rel_helperfunctions.contextual_predicate import process_data
from querent.kg.ner_helperfunctions.fixed_entities import FixedEntityExtractor
from querent.kg.ner_helperfunctions.ner_llm_transformer import NER_LLM
from querent.common.types.querent_event import EventState, EventType
from querent.core.base_engine import BaseEngine
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_messages import IngestedMessages
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.querent_queue import QuerentQueue
from querent.common.types.file_buffer import FileBuffer
from querent.kg.rel_helperfunctions.filter_semantic_triples import SemanticTripleFilter
from querent.logging.logger import setup_logger
from querent.kg.querent_kg import QuerentKG
from querent.config.graph_config import GraphConfig
from querent.kg.ner_helperfunctions.filter_triples import TripleFilter
from querent.config.core.llm_config import LLM_Config
from querent.kg.rel_helperfunctions.triple_to_json import TripleToJsonConverter
from querent.kg.rel_helperfunctions.embedding_store import EmbeddingStore
import time



class Fixed_Entities_LLM(BaseEngine):
    def __init__(
        self,
        input_queue:QuerentQueue,
        config: LLM_Config
    ):  
        self.logger = setup_logger(__name__, "Fixed_Entities_LLM")
        super().__init__(input_queue)
        self.skip_inferences=config.skip_inferences
        if not self.skip_inferences:
            mock_config = Opensource_LLM_Config(qa_template=config.user_context,
                                                model_type = config.rel_model_type,
                                                model_path = config.rel_model_path,
                                                grammar_file_path = config.grammar_file_path,
                                                emb_model_name = config.emb_model_name,
                                                is_confined_search = config.is_confined_search,
                                                spacy_model_path = config.spacy_model_path,
                                                nltk_path = config.nltk_path,
                                                )
            self.semantic_extractor = RelationExtractor(mock_config)
        self.graph_config = GraphConfig(identifier=config.name)
        self.contextual_graph = QuerentKG(self.graph_config)
        self.semantic_graph = QuerentKG(self.graph_config)
        self.file_buffer = FileBuffer()
        self.ner_tokenizer = AutoTokenizer.from_pretrained(config.ner_model_name)
        self.ner_llm_instance = NER_LLM(provided_tokenizer=self.ner_tokenizer, provided_model= "dummy")
        self.nlp_model = NER_LLM.set_nlp_model(config.spacy_model_path)
        self.nlp_model = NER_LLM.get_class_variable()
        huggingface_token = config.huggingface_token
        self.create_emb = EmbeddingStore(inference_api_key=huggingface_token)
        self.enable_filtering = config.enable_filtering
        self.filter_params = config.filter_params or {}
        self.triple_filter = None
        if self.enable_filtering:
            self.triple_filter = TripleFilter(**self.filter_params)
        self.sample_entities = config.sample_entities
        self.fixed_entities = config.fixed_entities
        if self.fixed_entities and not self.sample_entities:
            raise ValueError("If specific entities are provided, their types should also be provided.")
        if self.fixed_entities and self.sample_entities:
            self.entity_context_extractor = FixedEntityExtractor(fixed_entities=self.fixed_entities, entity_types=self.sample_entities,model = self.nlp_model)
        elif self.sample_entities:
            self.entity_context_extractor = FixedEntityExtractor(entity_types=self.sample_entities, model = self.nlp_model)
        else:
            self.entity_context_extractor = None
        self.fixed_relationships = config.fixed_relationships
        self.sample_relationships = config.sample_relationships
        if self.fixed_relationships and not self.sample_relationships:
            raise ValueError("If specific predicates are provided, their types should also be provided.")
        if self.fixed_relationships and self.sample_relationships:
            self.predicate_context_extractor = FixedPredicateExtractor(fixed_predicates=self.fixed_relationships, predicate_types=self.sample_relationships,model = self.nlp_model)
        elif self.sample_relationships:
            self.predicate_context_extractor = FixedPredicateExtractor(predicate_types=self.sample_relationships,model = self.nlp_model)
        else:
            self.predicate_context_extractor = None
        self.user_context = config.user_context
        self.isConfinedSearch = config.is_confined_search
        self.semantictriplefilter = SemanticTripleFilter()
        
 

    def validate(self) -> bool:
        return self.ner_tokenizer is not None

    def process_messages(self, data: IngestedMessages):
        return super().process_messages(data)
    
    def process_images(self, data: IngestedImages):
        return super().process_images(data)

    async def process_code(self, data: IngestedCode):
        return super().process_code(data)

    @staticmethod
    def validate_ingested_tokens(data: IngestedTokens) -> bool:
        if data.is_error():
            
            return False

        return True

    def count_entity_pairs(self, doc_entity_pairs):
        total_pairs = 0
        for inner_list in doc_entity_pairs:
            total_pairs += len(inner_list)
            
        return total_pairs
    
    def set_filter_params(self, **kwargs):
        self.filter_params = kwargs
        if self.triple_filter:
            self.triple_filter.set_params(**kwargs)
        else:
            self.triple_filter = TripleFilter(**kwargs)
    
    
    async def process_tokens(self, data: IngestedTokens):
        doc_entity_pairs = []
        number_sentences = 0
        try:
            if not Fixed_Entities_LLM.validate_ingested_tokens(data):
                    self.set_termination_event()                                      
                    return
            if data.data:
                single_string = ' '.join(data.data)
                clean_text = single_string
                # clean_text = unidecode(single_string)
            else:
                clean_text = data.data
            if not data.is_token_stream : 
                file, content = self.file_buffer.add_chunk(
                data.get_file_path(), clean_text)
            else:
                content = clean_text
                file = data.get_file_path()
            if content:
                if self.fixed_entities:
                    content = self.entity_context_extractor.find_entity_sentences(content)
                if self.fixed_relationships:
                    content = self.predicate_context_extractor.find_predicate_sentences(content)
                tokens = self.ner_llm_instance._tokenize_and_chunk(content)
                for tokenized_sentence, original_sentence, sentence_idx in tokens:
                    (entities, entity_pairs,) = self.ner_llm_instance.extract_entities_from_sentence(original_sentence, sentence_idx, [s[1] for s in tokens],self.isConfinedSearch, self.fixed_entities, self.sample_entities)
                    if entity_pairs:
                        doc_entity_pairs.append(self.ner_llm_instance.transform_entity_pairs(entity_pairs))
                    number_sentences = number_sentences + 1
            else:
                return

            if self.sample_entities:
                doc_entity_pairs = self.entity_context_extractor.process_entity_types(doc_entities=doc_entity_pairs)
            if doc_entity_pairs and any(doc_entity_pairs):
                doc_entity_pairs = self.ner_llm_instance.remove_duplicates(doc_entity_pairs)
                filtered_triples = process_data(doc_entity_pairs, file)
                if not filtered_triples:
                    self.logger.debug("No entity pairs")
                    return
                elif not self.skip_inferences:
                    relationships = self.semantic_extractor.process_tokens(filtered_triples)
                    self.logger.debug(f"length of relationships {len(relationships)}")
                    relationships = self.semantictriplefilter.filter_triples(relationships)
                    if len(relationships) > 0:
                        embedding_triples = self.create_emb.generate_embeddings(relationships)
                        if self.sample_relationships:
                            embedding_triples = self.predicate_context_extractor.process_predicate_types(embedding_triples)
                        for triple in embedding_triples:
                            graph_json = json.dumps(TripleToJsonConverter.convert_graphjson(triple))
                            if graph_json:
                                current_state = EventState(EventType.Graph,1.0, graph_json, file)
                                await self.set_state(new_state=current_state)
                            vector_json = json.dumps(TripleToJsonConverter.convert_vectorjson(triple))
                            if vector_json:
                                current_state = EventState(EventType.Vector,1.0, vector_json, file)
                                await self.set_state(new_state=current_state)
                    else:
                        return
                else:
                    return filtered_triples, file
            else:
                return
        except Exception as e:
            self.logger.debug(f"Invalid {self.__class__.__name__} configuration. Unable to process tokens. {e}")



from pydantic import BaseModel, Field
from typing import Dict
import os

class Opensource_LLM_Config(BaseModel):
    name: str = "RelationshipExtractor"
    description: str = "An engine for extracting relationships"
    version: str = "0.0.1"
    logger: str = "RelationshipExtractor.engine_config"
    model_type: str = 'llama'
    model_path: str = './tests/llama-2-7b-chat.Q5_K_M.gguf' 
    grammar_file_path: str = './querent/kg/rel_helperfunctions/json.gbnf'
    qa_template: str = Field(default=None)
    vector_store_path: str = "./querent/kg/rel_helperfunctions/vectorstores/"
    emb_model_name: str = 'sentence-transformers/all-MiniLM-L6-v2'
    rag_approach: bool = False
    is_confined_search: bool = False
    spacy_model_path: str = 'en_core_web_lg'
    nltk_path: str = '/model/nltk_data'

    def __init__(self, config_source=None, **kwargs):
        config_data = {}
        config_data.update(kwargs)
        if config_source:
            config_data = self.load_config(config_source)
        if "config" in config_data:
            config_data.update(config_data["config"])
        super().__init__(**config_data)
        

    @classmethod
    def load_config(cls, config_source) -> dict:
        if isinstance(config_source, dict):
            # If config source is a dictionary, return a dictionary
            cls.config_data = config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

        env_vars = dict(os.environ)
        cls.config_data.update(env_vars)
        return cls.config_data

    def get_faiss_index_path(self):
        return self.vector_store_path + "my_FAISS_index"


from typing import List, AsyncGenerator
import csv
import io

from querent.processors.async_processor import AsyncProcessor
from querent.ingestors.ingestor_factory import IngestorFactory
from querent.ingestors.base_ingestor import BaseIngestor
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.common.types.collected_bytes import CollectedBytes
from querent.common import common_errors
from querent.common.types.ingested_tokens import IngestedTokens
from querent.logging.logger import setup_logger


class CsvIngestorFactory(IngestorFactory):
    SUPPORTED_EXTENSIONS = {"csv"}

    async def supports(self, file_extension: str) -> bool:
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        if not await self.supports(file_extension):
            return None
        return CsvIngestor(processors)


class CsvIngestor(BaseIngestor):
    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.CSV)
        self.processors = processors
        self.logger = setup_logger(__name__, "CsvIngestor")

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens, None]:
        current_file = None
        collected_bytes = b""
        try:
            async for chunk_bytes in poll_function:
                if chunk_bytes.is_error() or chunk_bytes.is_eof():
                    # TODO handle error
                    continue
                if current_file is None:
                    current_file = chunk_bytes.file
                elif current_file != chunk_bytes.file:
                    # we have a new file, process the old one
                    async for row in self.extract_and_process_csv(
                        CollectedBytes(file=current_file, data=collected_bytes)
                    ):
                        yield IngestedTokens(file=current_file, data=[row], error=None)
                    yield IngestedTokens(
                        file=current_file,
                        data=None,
                        error=None,
                    )
                    collected_bytes = b""
                    current_file = chunk_bytes.file
                collected_bytes += chunk_bytes.data
        except Exception as e:
            yield IngestedTokens(file=current_file, data=None, error=f"Exception: {e}")
        finally:
            # process the last file
            async for row in self.extract_and_process_csv(
                CollectedBytes(file=current_file, data=collected_bytes)
            ):
                yield IngestedTokens(file=current_file, data=[row], error=None)

            yield IngestedTokens(file=current_file, data=None, error=None)

    async def extract_and_process_csv(
        self, collected_bytes: CollectedBytes
    ) -> AsyncGenerator[str, None]:
        text = await self.extract_text_from_csv(collected_bytes)
        rows = text.splitlines()
        for row in rows:
            processed_row = await self.process_data(row)
            yield processed_row

    async def extract_text_from_csv(self, collected_bytes: CollectedBytes) -> str:
        try:
            text_data = collected_bytes.data.decode("utf-8")
            return text_data
        except UnicodeDecodeError as exc:
            raise common_errors.UnicodeDecodeError(
                f"Getting UnicodeDecodeError on this file {collected_bytes.file}"
            ) from exc
        except LookupError as exc:
            raise common_errors.LookupError(
                f"Getting LookupError on this file {collected_bytes.file}"
            ) from exc
        except TypeError as exc:
            raise common_errors.TypeError(
                f"Getting TypeError on this file {collected_bytes.file}"
            ) from exc
        except Exception as exc:
            raise common_errors.UnknownError(
                f"Getting error message:- {exc} on file {collected_bytes.file}"
            )

    async def process_data(self, text: str) -> str:
        if self.processors is None or len(self.processors) == 0:
            return text
        try:
            processed_data = text
            for processor in self.processors:
                processed_data = await processor.process_text(processed_data)
            return processed_data
        except Exception as e:
            self.logger.error(f"Error while processing text: {e}")



# import time

# import numpy as np
# from pymilvus import (
#     connections,
#     utility,
#     FieldSchema, CollectionSchema, DataType,
#     Collection,
# )

# class MilvusDBConnection:
    
#     def create_collection(self,collection_name, dim):
#         # if utility.has_collection(collection_name):
#         #     utility.drop_collection(collection_name)
#         connections.connect("default", host="localhost", port="19530")
#         has = utility.has_collection(collection_name)
#         fields = [
#         FieldSchema(name='knowledge', dtype=DataType.VARCHAR, descrition='ids (subject_predicate_object)', max_length=500, is_primary=True, auto_id=False),
#         FieldSchema(name='relationship', dtype=DataType.VARCHAR, max_length=500, descrition='namespace',default_value="Unknown"),
#         FieldSchema(name='embeddings', dtype=DataType.FLOAT_VECTOR, descrition='embedding vectors', dim=dim),
#         FieldSchema(name='document', dtype=DataType.VARCHAR, max_length=500, default_value="Unknown")
#         ]
#         schema = CollectionSchema(fields=fields, description='semantic search')
#         if not utility.has_collection(collection_name):
#             collection = Collection(name=collection_name, schema=schema)
#             # create IVF_FLAT index for collection.
#             # index_params = {
#             #     'metric_type':'L2',
#             #     'index_type':"IVF_FLAT",
#             #     'params':{"nlist":2048}
#             # }
#             # collection.create_index(field_name="embedding", index_params=index_params)
#         else:
#             collection = Collection(name=collection_name)

#         return collection
    
#     def insert_vector_event(self, id, embedding, namespace, document, collection):
#         if collection is not None:
#             try:
#                 insert_result = collection.insert(
#                                                     [
#                                                         [id],
#                                                         [namespace],
#                                                         [embedding],
#                                                         [document] 
                                                        
#                                                     ]
#                                                 )
                
#             except Exception as e:
#                 raise("An error occurred while inserting the data:")

#     def vectorsearch_db(self, collection, vector_search_query):
#         collection.load()
#         vectors_to_search = vector_search_query
#         search_params = {
#             "metric_type": "L2",
#             "params": {"nprobe": 10},
#         }

#         start_time = time.time()
#         result = collection.search(vectors_to_search, "embeddings", search_params, limit=3)
#         end_time = time.time()

#         return result
        


from typing import Optional
from querent.config.core.llm_config import LLM_Config
import os

class GPTConfig(LLM_Config):
    id: str = ""
    name: str = "OPENAIEngine"
    description: str = "An engine for NER using BERT and knowledge graph operations using OPENAI"
    version: str = "0.0.1"
    logger: str = "OPENAI.engine_config"
    ner_model_name: str = "dbmdz/bert-large-cased-finetuned-conll03-english"
    rel_model_name: str = "gpt-3.5-turbo"
    requests_per_minute: int = 3
    openai_api_key: str = ""
    user_context: str = None
    huggingface_token: Optional[str] = None

    def __init__(self, config_source=None, **kwargs):
        config_data = {}
        config_data.update(kwargs)
        if config_source:
            config_data = self.load_config(config_source)
        if "config" in config_data:
            config_data.update(config_data["config"])
        super().__init__(**config_data)

    
    @classmethod
    def load_config(cls, config_source) -> dict:
        if isinstance(config_source, dict):
            # If config source is a dictionary, return a dictionary
            cls.config_data = config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

        env_vars = dict(os.environ)
        cls.config_data.update(env_vars)
        return cls.config_data


import asyncio
import json
import re
from querent.core.transformers.fixed_entities_set_opensourcellm import Fixed_Entities_LLM
from querent.kg.ner_helperfunctions.fixed_predicate import FixedPredicateExtractor
from querent.config.core.gpt_llm_config import GPTConfig
from querent.core.transformers.bert_ner_opensourcellm import BERTLLM
from querent.common.types.ingested_images import IngestedImages
from querent.kg.ner_helperfunctions.ner_llm_transformer import NER_LLM
from querent.kg.rel_helperfunctions.openai_functions import FunctionRegistry
from querent.common.types.querent_event import EventState, EventType
from querent.core.base_engine import BaseEngine
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_messages import IngestedMessages
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.querent_queue import QuerentQueue
from querent.kg.rel_helperfunctions.embedding_store import EmbeddingStore
from typing import Any, List, Tuple
from querent.kg.rel_helperfunctions.triple_to_json import TripleToJsonConverter
from querent.logging.logger import setup_logger
from querent.config.core.llm_config import LLM_Config
from openai import OpenAI
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    wait_fixed
)
from dotenv import load_dotenv, find_dotenv
import json

_ = load_dotenv(find_dotenv())


class GPTLLM(BaseEngine):
    def __init__(
        self,
        input_queue: QuerentQueue,
        config: GPTConfig
    ):
        self.logger = setup_logger(__name__, "OPENAILLM")
        try:
            super().__init__(input_queue)
            llm_config = LLM_Config(
            ner_model_name=config.ner_model_name,
            enable_filtering=config.enable_filtering,
            filter_params={
                'score_threshold': config.filter_params['score_threshold'],
                'attention_score_threshold': config.filter_params['attention_score_threshold'],
                'similarity_threshold': config.filter_params['similarity_threshold'],
                'min_cluster_size': config.filter_params['min_cluster_size'],
                'min_samples': config.filter_params['min_samples'],
                'cluster_persistence_threshold':config.filter_params['cluster_persistence_threshold']
            },
            sample_entities = config.sample_entities,
            fixed_entities = config.fixed_entities,
            skip_inferences= True,
            is_confined_search = config.is_confined_search,
            huggingface_token = config.huggingface_token,
            spacy_model_path = config.spacy_model_path,
            nltk_path = config.nltk_path)
            self.fixed_entities = config.fixed_entities
            self.is_confined_search = config.is_confined_search
            self.fixed_relationships = config.fixed_relationships
            self.sample_relationships = config.sample_relationships
            self.user_context = config.user_context
            self.nlp_model = NER_LLM.set_nlp_model(config.spacy_model_path)
            self.nlp_model = NER_LLM.get_class_variable()
            if self.fixed_relationships and not self.sample_relationships:
                raise ValueError("If specific predicates are provided, their types should also be provided.")
            if self.fixed_relationships and self.sample_relationships:
                self.predicate_context_extractor = FixedPredicateExtractor(fixed_predicates=self.fixed_relationships, predicate_types=self.sample_relationships,  model = self.nlp_model)
            elif self.sample_relationships:
                self.predicate_context_extractor = FixedPredicateExtractor(predicate_types=self.sample_relationships,  model = self.nlp_model)
            else:
                self.predicate_context_extractor = None
            self.create_emb = EmbeddingStore(inference_api_key=config.huggingface_token)
            if config.is_confined_search:
                self.llm_instance = Fixed_Entities_LLM(input_queue, llm_config)
            else :
                self.llm_instance = BERTLLM(input_queue, llm_config)
            self.rel_model_name = config.rel_model_name
            if config.openai_api_key:
                self.gpt_llm = OpenAI(api_key=config.openai_api_key)
            else:
                self.gpt_llm = OpenAI()
            self.function_registry = FunctionRegistry()
            
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to Initialize. {e}")
            raise Exception(f"Invalid {self.__class__.__name__} configuration. Unable to Initialize. {e}")
    
    def validate(self) -> bool:
        return isinstance(self.llm_instance, BERTLLM) or isinstance(self.llm_instance, Fixed_Entities_LLM)

    def process_messages(self, data: IngestedMessages):
        return super().process_messages(data)
    
    def process_images(self, data: IngestedImages):
        return super().process_messages(data)

    async def process_code(self, data: IngestedCode):
        return super().process_messages(data)

    @staticmethod
    def validate_ingested_tokens(data: IngestedTokens) -> bool:
        if data.is_error():
            
            return False

        return True
    def extract_semantic_triples(self, chat_completion):
    # Extract the message content from the ChatCompletion
        message_content = chat_completion.choices[0].message.content.replace('\n', '')

        # Parse the JSON content into a Python list
        try:
            triples_list = eval(message_content)
            if not isinstance(triples_list, list):
                raise ValueError("Content is not a list")
        except Exception as e:
            return []

        return triples_list

    @staticmethod
    def remove_items_from_tuples(data: List[Tuple[str, str, str]]) -> List[Tuple[str, str, str]]:
        result = []
        keys_to_remove = ['entity1_embedding', 'entity2_embedding', 'entity1_attnscore', 'entity2_attnscore', 'pair_attnscore']

        for tup in data:
            json_data = json.loads(tup[1])
            for key in keys_to_remove:
                json_data.pop(key, None)
            modified_json_str = json.dumps(json_data, ensure_ascii=False)
            modified_tuple = (tup[0], modified_json_str, tup[2])
            result.append(modified_tuple)

        return result
    
    async def process_triples(self, context, entity1, entity2, entity1_label, entity2_label):
        try: 
            if not self.user_context:
                identify_entity_message = f"""Please analyze the provided context below. Once you have understood the context, answer the user query using the specified output format.
        
                    Context: {context}
                    Entity 1: {entity1} and Entity 2: {entity2}
                    Output Format:
                    [
                        {{
                            'subject': 'Identified as the main entity in the context, typically the initiator or primary focus of the action or topic being discussed.',
                            'predicate': 'The relationship (predicate) between the subject and the object.',
                            'object': 'This parameter represents the entity in the context directly impacted by or involved in the action, typically the recipient or target of the main verb's action.',
                            'subject_type': 'The category of the subject entity e.g. location, person, event, material, process etc.',
                            'object_type': 'The category of the object entity e.g. location, person, event, material, process etc.',
                            'predicate_type': 'The category of the predicate e.g. causative, action, ownership, occurance etc.'
                        }},
                    ]
                    """
                messages_classify_entity = [
                    {"role": "user", "content": identify_entity_message},
                    {"role": "user", "content": "Query: First, identify all geological entities in the provided context. Then, create relevant semantic triples (Subject, Predicate, Object) and also categorize the respective the Subject, Object types (e.g. location, person, event, material, process etc.) and Predicate type. Use the above output format to provide all the relevant semantic triples."},
                ]
            elif self.user_context and self.fixed_entities :
                identify_entity_message = f"""Please analyze the provided context below. Once you have understood the context, answer the user query using the specified output format.
        
                    Context: {context}
                    Entity 1: {entity1} and Entity 2: {entity2}
                    Entity 1_Type: {entity1_label} and Entity 2_Type: {entity2_label}
                    Output Format:
                    [
                        {{
                            'subject': 'Identified as the main entity in the context, typically the initiator or primary focus of the action or topic being discussed.',
                            'predicate': 'The relationship (predicate) between the subject and the object.',
                            'object': 'This parameter represents the entity in the context directly impacted by or involved in the action, typically the recipient or target of the main verb's action.',
                            'subject_type': 'The category of the subject entity e.g. location, person, event, material, process etc.',
                            'object_type': 'The category of the object entity e.g. location, person, event, material, process etc.',
                            'predicate_type': 'The category of the predicate e.g. causative, action, ownership, occurance etc.'
                        }},
                    ]
                    """
                messages_classify_entity = [
                    {"role": "user", "content": identify_entity_message},
                    {"role": "user", "content": self.user_context},
                ]
            elif self.user_context and not self.fixed_entities :
                identify_entity_message = f"""Please analyze the provided context below. Once you have understood the context, answer the user query using the specified output format.
        
                    Context: {context}
                    Entity 1: {entity1} and Entity 2: {entity2}
                    Output Format:
                    [
                        {{
                            'subject': 'Identified as the main entity in the context, typically the initiator or primary focus of the action or topic being discussed.',
                            'predicate': 'The relationship (predicate) between the subject and the object.',
                            'object': 'This parameter represents the entity in the context directly impacted by or involved in the action, typically the recipient or target of the main verb's action.',
                            'subject_type': 'The category of the subject entity e.g. location, person, event, material, process etc.',
                            'object_type': 'The category of the object entity e.g. location, person, event, material, process etc.',
                            'predicate_type': 'The category of the predicate e.g. causative, action, ownership, occurance etc.'
                        }},
                    ]
                    """
                messages_classify_entity = [
                    {"role": "user", "content": identify_entity_message},
                    {"role": "user", "content": self.user_context},
                ]
            identify_predicate_response = self.generate_response(
                messages_classify_entity,
                "predicate_info"
            )
            semantic_triples = self.extract_semantic_triples(identify_predicate_response)
            if len(semantic_triples)>0:
                return {
                    'subject_type': semantic_triples[0]['subject_type'].lower().replace(" ", "_"),
                    'subject': semantic_triples[0]['subject'].lower(),
                    'object_type': semantic_triples[0]['object_type'].lower().replace(" ", "_"),
                    'object': semantic_triples[0]['object'].lower(),
                    'predicate': semantic_triples[0]['predicate'].lower(),
                    'predicate_type': semantic_triples[0]['predicate_type'].lower().replace(" ", "_")
                }   
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to process triples using GPT. {e}")

    # @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6))
    def completion_with_backoff(self, **kwargs):
        return self.gpt_llm.chat.completions.create(**kwargs)
    
    def generate_response(self, messages, name):
        response = self.completion_with_backoff(
            model=self.rel_model_name,
            messages=messages,
            temperature=0
        )
        return response

    def generate_output_tuple(self,result, context_json):
        context_data = json.loads(context_json)
        context = context_data.get("context", "")
        subject_type = result.get("subject_type", "Unlabeled")
        subject = result.get("subject", "")
        object_type = result.get("object_type", "Unlabeled")
        object = result.get("object", "")
        predicate = result.get("predicate", "")
        predicate_type = result.get("predicate_type", "Unlabled")
        output_tuple = (
            subject,
            f'{{"predicate": "{predicate}", "predicate_type": "{predicate_type}", "context": "{context}", "file": "{context_data.get("file_path", "")}", "subject_type": "{subject_type}", "object_type": "{object_type}"}}',
            object
        )

        return output_tuple
      
    async def process_tokens(self, data: IngestedTokens):
        try:
            if not GPTLLM.validate_ingested_tokens(data):
                    self.set_termination_event()                    
                    return 
            relationships = []
            result = await self.llm_instance.process_tokens(data)           
            if not result: return 
            else:
                filtered_triples, file = result
                modified_data = GPTLLM.remove_items_from_tuples(filtered_triples)
                for entity1, context_json, entity2 in modified_data:
                    context_data = json.loads(context_json)
                    context = context_data.get("context", "")
                    entity1_label = context_data.get("entity1_label", "")
                    entity2_label = context_data.get("entity2_label", "")
                    entity1_nn_chunk = context_data.get("entity1_nn_chunk","")
                    entity2_nn_chunk = context_data.get("entity2_nn_chunk","")
                    result = await self.process_triples(context, entity1_nn_chunk, entity2_nn_chunk, entity1_label, entity2_label)
                    if result:
                        output_tuple = self.generate_output_tuple(result, context_json)
                        relationships.append(output_tuple)
                if len(relationships) > 0:
                    embedding_triples = self.create_emb.generate_embeddings(relationships)
                    if self.sample_relationships:
                            embedding_triples = self.predicate_context_extractor.process_predicate_types(embedding_triples)
                    for triple in embedding_triples:
                        graph_json = json.dumps(TripleToJsonConverter.convert_graphjson(triple))
                        if graph_json:
                                current_state = EventState(EventType.Graph,1.0, graph_json, file)
                                await self.set_state(new_state=current_state)
                        vector_json = json.dumps(TripleToJsonConverter.convert_vectorjson(triple))
                        if vector_json:
                                current_state = EventState(EventType.Vector,1.0, vector_json, file)
                                await self.set_state(new_state=current_state)
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to extract predicates using GPT. {e}")
            raise Exception(f"An error occurred while extracting predicates using GPT: {e}")

    async def process_messages(self, data: IngestedMessages):
        raise NotImplementedError